@@ -62,197 +62,265 @@ use std::{
 use anyhow::{Context, Result};
 
 use chrono::{DateTime, NaiveTime, offset::Utc};
-use structopt::{StructOpt, clap::AppSettings};
+use clap::{CommandFactory, FromArgMatches, Parser};
 
 use crate::{
     ArmorLabel,
     EncryptAs,
+    Save,
+    Load,
     InlineSignAs,
     Password,
     SignAs,
+    SOP,
 };
 
-#[derive(Debug, StructOpt)]
-#[structopt(about = "An implementation of the \
-                     Stateless OpenPGP Command Line Interface",
-            settings(&[AppSettings::DisableVersion,
-                       AppSettings::VersionlessSubcommands]),
-            )]
+#[derive(Debug, Parser)]
+#[clap(about = "An implementation of the Stateless OpenPGP Command Line Interface")]
+#[clap(disable_version_flag(true))]
 enum Operation {
     /// Prints version information.
     ///
     /// Invoked without arguments, returns name and version of the SOP
     /// implementation.
-    #[structopt(display_order = 100)]
+    #[clap(display_order = 100)]
     Version {
         /// Returns name and version of the primary underlying OpenPGP
         /// toolkit.
-        #[structopt(long, conflicts_with("extended"))]
+        #[clap(long, conflicts_with("extended"), conflicts_with("sop_spec"))]
         backend: bool,
 
         /// Returns multiple lines of name and version information.
         ///
         /// The first line is the name and version of the SOP
         /// implementation, but the rest have no defined structure.
-        #[structopt(long, conflicts_with("backend"))]
+        #[clap(long, conflicts_with("backend"), conflicts_with("sop_spec"))]
         extended: bool,
+
+        /// Returns the latest version of the SOP spec that is
+        /// implemented.
+        #[clap(long, conflicts_with("backend"), conflicts_with("extended"))]
+        sop_spec: bool,
+    },
+    /// Emits a list of profiles supported by the identified
+    /// subcommand.
+    #[clap(display_order = 150)]
+    ListProfiles {
+        subcommand: String,
     },
     /// Generates a Secret Key.
-    #[structopt(display_order = 200)]
+    #[clap(display_order = 200)]
     GenerateKey {
         /// Don't ASCII-armor output.
-        #[structopt(long)]
+        #[clap(long)]
         no_armor: bool,
+        /// Select the profile to use for key generation.
+        #[clap(long)]
+        profile: Option<String>,
+        /// Create a signing-only key.
+        #[clap(long)]
+        signing_only: bool,
         /// Protect the newly generated key with the given password.
-        #[structopt(long)]
+        #[clap(long)]
         with_key_password: Option<String>,
         /// UserIDs for the generated key.
         userids: Vec<String>,
     },
+
+    /// Updates a key's password.
+    ///
+    /// The output will be the same set of OpenPGP Transferable Secret
+    /// Keys as the input, but with all secret key material locked
+    /// according to the password indicated by the
+    /// `--new-key-password` option (or, with no password at all, if
+    /// `--new-key-password` is absent).  Note that
+    /// `--old-key-password` can be supplied multiple times, and each
+    /// supplied password will be tried as a means to unlock any
+    /// locked key material encountered.  It will normalize a
+    /// Transferable Secret Key to use a single password even if it
+    /// originally had distinct passwords locking each of the subkeys.
+    ///
+    /// If any secret key packet is locked but cannot be unlocked with
+    /// any of the supplied `--old-key-password` arguments, this
+    /// subcommand should fail with `KEY_IS_PROTECTED`.
+    #[clap(display_order = 210)]
+    ChangeKeyPassword {
+        /// Don't ASCII-armor output.
+        #[clap(long)]
+        no_armor: bool,
+        /// The new password to lock the key with, or just unlock the
+        /// key if the option is absent.
+        #[clap(long)]
+        new_key_password: Option<String>,
+        /// Unlock the keys with these passwords.
+        #[clap(long, number_of_values = 1)]
+        old_key_password: Vec<String>,
+    },
+
+    /// Creates a Revocation Certificate.
+    ///
+    /// Generate a revocation certificate for each Transferable Secret
+    /// Key found.
+    #[clap(display_order = 220)]
+    RevokeKey {
+        /// Don't ASCII-armor output.
+        #[clap(long)]
+        no_armor: bool,
+        /// Unlock the keys with these passwords.
+        #[clap(long, number_of_values = 1)]
+        with_key_password: Vec<String>,
+    },
+
     /// Extracts a Certificate from a Secret Key.
-    #[structopt(display_order = 300)]
+    #[clap(display_order = 300)]
     ExtractCert {
         /// Don't ASCII-armor output.
-        #[structopt(long)]
+        #[clap(long)]
         no_armor: bool,
     },
     /// Creates Detached Signatures.
-    #[structopt(display_order = 400)]
+    #[clap(display_order = 400)]
     Sign {
         /// Don't ASCII-armor output.
-        #[structopt(long)]
+        #[clap(long)]
         no_armor: bool,
         /// Sign binary data or UTF-8 text.
-        #[structopt(default_value = "binary", long = "as")]
+        #[clap(default_value = "binary", long = "as")]
         as_: SignAs,
         /// Emit the digest algorithm used to the specified file.
-        #[structopt(long)]
+        #[clap(long)]
         micalg_out: Option<String>,
         /// Try to decrypt the signing KEYS with these passwords.
-        #[structopt(long, number_of_values = 1)]
+        #[clap(long, number_of_values = 1)]
         with_key_password: Vec<String>,
         /// Keys for signing.
         keys: Vec<String>,
     },
     /// Verifies Detached Signatures.
-    #[structopt(display_order = 500)]
+    #[clap(display_order = 500)]
     Verify {
         /// Consider signatures before this date invalid.
-        #[structopt(long, parse(try_from_str = parse_bound_round_down))]
-        not_before: Option<DateTime<Utc>>,
+        #[clap(long)]
+        not_before: Option<NotBefore>,
         /// Consider signatures after this date invalid.
-        #[structopt(long, parse(try_from_str = parse_bound_round_up))]
-        not_after: Option<DateTime<Utc>>,
+        #[clap(long)]
+        not_after: Option<NotAfter>,
         /// Signatures to verify.
         signatures: String,
         /// Certs for verification.
         certs: Vec<String>,
     },
     /// Encrypts a Message.
-    #[structopt(display_order = 600)]
+    #[clap(display_order = 600)]
     Encrypt {
         /// Don't ASCII-armor output.
-        #[structopt(long)]
+        #[clap(long)]
         no_armor: bool,
+        /// Select the profile to use for encryption.
+        #[clap(long)]
+        profile: Option<String>,
         /// Encrypt binary data or UTF-8 text.
-        #[structopt(default_value = "binary", long = "as")]
+        #[clap(default_value = "binary", long = "as")]
         as_: EncryptAs,
         /// Encrypt with passwords.
-        #[structopt(long, number_of_values = 1)]
+        #[clap(long, number_of_values = 1)]
         with_password: Vec<String>,
         /// Keys for signing.
-        #[structopt(long, number_of_values = 1)]
+        #[clap(long, number_of_values = 1)]
         sign_with: Vec<String>,
         /// Try to decrypt the signing KEYS with these passwords.
-        #[structopt(long, number_of_values = 1)]
+        #[clap(long, number_of_values = 1)]
         with_key_password: Vec<String>,
+        /// Write the session key here.
+        #[clap(long)]
+        session_key_out: Option<String>,
         /// Encrypt for these certs.
         certs: Vec<String>,
     },
     /// Decrypts a Message.
-    #[structopt(display_order = 700)]
+    #[clap(display_order = 700)]
     Decrypt {
         /// Write the session key here.
-        #[structopt(long)]
+        #[clap(long)]
         session_key_out: Option<String>,
         /// Try to decrypt with this session key.
-        #[structopt(long, number_of_values = 1)]
+        #[clap(long, number_of_values = 1)]
         with_session_key: Vec<String>,
         /// Try to decrypt with this password.
-        #[structopt(long, number_of_values = 1)]
+        #[clap(long, number_of_values = 1)]
         with_password: Vec<String>,
         /// Write verification result here.
-        #[structopt(long, alias("verify-out"))]
+        #[clap(long, alias("verify-out"))]
         verifications_out: Option<String>,
         /// Certs for verification.
-        #[structopt(long, number_of_values = 1)]
+        #[clap(long, number_of_values = 1)]
         verify_with: Vec<String>,
         /// Consider signatures before this date invalid.
-        #[structopt(long, parse(try_from_str = parse_bound_round_down))]
-        verify_not_before: Option<DateTime<Utc>>,
+        #[clap(long)]
+        verify_not_before: Option<NotBefore>,
         /// Consider signatures after this date invalid.
-        #[structopt(long, parse(try_from_str = parse_bound_round_up))]
-        verify_not_after: Option<DateTime<Utc>>,
+        #[clap(long)]
+        verify_not_after: Option<NotAfter>,
         /// Try to decrypt the encryption KEYS with these passwords.
-        #[structopt(long, number_of_values = 1)]
+        #[clap(long, number_of_values = 1)]
         with_key_password: Vec<String>,
         /// Try to decrypt with these keys.
         keys: Vec<String>,
     },
     /// Converts binary OpenPGP data to ASCII.
-    #[structopt(display_order = 800)]
+    #[clap(display_order = 800)]
     Armor {
         /// Indicates the kind of data.
-        #[structopt(long, default_value = "auto")]
+        #[clap(long, default_value = "auto", hide(true))]
         label: ArmorLabel,
     },
     /// Converts ASCII OpenPGP data to binary.
-    #[structopt(display_order = 900)]
+    #[clap(display_order = 900)]
     Dearmor {
     },
     /// Splits Signatures from an Inline-Signed Message.
-    #[structopt(display_order = 950)]
+    #[clap(display_order = 750)]
     InlineDetach {
         /// Don't ASCII-armor the signatures.
-        #[structopt(long)]
+        #[clap(long)]
         no_armor: bool,
         /// Write Signatures here.
-        #[structopt(long)]
+        #[clap(long)]
         signatures_out: String,
     },
     /// Verifies Inline-Signed Messages.
-    #[structopt(display_order = 951)]
+    #[clap(display_order = 751)]
     InlineVerify {
         /// Consider signatures before this date invalid.
-        #[structopt(long, parse(try_from_str = parse_bound_round_down))]
-        not_before: Option<DateTime<Utc>>,
+        #[clap(long)]
+        not_before: Option<NotBefore>,
         /// Consider signatures after this date invalid.
-        #[structopt(long, parse(try_from_str = parse_bound_round_up))]
-        not_after: Option<DateTime<Utc>>,
+        #[clap(long)]
+        not_after: Option<NotAfter>,
         /// Write verification result here.
-        #[structopt(long)]
+        #[clap(long)]
         verifications_out: Option<String>,
         /// Certs for verification.
         certs: Vec<String>,
     },
     /// Creates Inline-Signed Messages.
-    #[structopt(display_order = 951)]
+    #[clap(display_order = 752)]
     InlineSign {
         /// Don't ASCII-armor output.
-        #[structopt(long)]
+        #[clap(long)]
         no_armor: bool,
         /// Sign binary data, UTF-8 text, or using the Cleartext
         /// Signature Framework.
-        #[structopt(default_value = "binary", long = "as")]
+        #[clap(default_value = "binary", long = "as")]
         as_: InlineSignAs,
         /// Try to decrypt the signing KEYS with these passwords.
-        #[structopt(long, number_of_values = 1)]
+        #[clap(long, number_of_values = 1)]
         with_key_password: Vec<String>,
         /// Keys for signing.
         keys: Vec<String>,
     },
     /// Unsupported subcommand.
-    #[structopt(external_subcommand)]
+    #[clap(external_subcommand)]
     Unsupported(Vec<String>),
 }
 
@@ -263,25 +331,32 @@ pub fn write_shell_completions<B, P>(binary_name: B,
 where B: AsRef<str>,
       P: AsRef<Path>,
 {
-    use structopt::clap::Shell;
+    use clap_complete::{generate_to, Shell};
 
     let binary_name = binary_name.as_ref();
     let out_path = out_path.as_ref();
     std::fs::create_dir_all(&out_path)?;
 
-    let mut clap = Operation::clap();
+    let mut clap = Operation::command();
     for shell in &[Shell::Bash, Shell::Fish, Shell::Zsh, Shell::PowerShell,
                    Shell::Elvish] {
-        clap.gen_completions(binary_name, *shell, out_path);
+        let path = generate_to(*shell, &mut clap, binary_name, out_path)?;
+        println!("cargo:warning=completion file is generated: {:?}", path);
     }
     Ok(())
 }
 
 /// Implements the SOP command-line interface.
-pub fn main(sop: &dyn crate::SOP) -> ! {
+pub fn main<'s, S, C, K, Sigs>(sop: &'s S) -> !
+where
+    S: SOP<'s, Certs = C, Keys = K, Sigs = Sigs>,
+    C: Load<'s, S> + Save,
+    K: Load<'s, S> + Save,
+    Sigs: Load<'s, S> + Save,
+{
     use std::process::exit;
 
-    match real_main(sop) {
+    match real_main::<S, C, K, Sigs>(sop) {
         Ok(()) => exit(0),
         Err(e) => {
             print_error_chain(&e);
@@ -302,48 +377,90 @@ pub fn main(sop: &dyn crate::SOP) -> ! {
     }
 }
 
-fn real_main(sop: &dyn crate::SOP) -> Result<()> {
-    // Do a little dance to supply name and version to structopt.
+fn real_main<'s, S, C, K, Sigs>(sop: &'s S) -> Result<()>
+where
+    S: SOP<'s, Certs = C, Keys = K, Sigs = Sigs>,
+    C: Load<'s, S> + Save,
+    K: Load<'s, S> + Save,
+    Sigs: Load<'s, S> + Save,
+{
+    // Do a little dance to supply name and version to clap.
     let v = sop.version()?.frontend()?;
+    let name = format!("SOP {}", v.name);
     let version = v.version.clone();
     let about = format!("An implementation of the Stateless OpenPGP Command \
                         Line Interface using {} {}",
                         v.name, v.version);
-    let app = Operation::clap()
-        .name(format!("SOP {}", v.name))
-        .about(&about[..])
-        .version(&version[..]);
+    let app = Operation::command()
+        .name(Box::leak(name.into_boxed_str()) as &str)
+        .about(Box::leak(about.into_boxed_str()) as &str)
+        .version(Box::leak(version.into_boxed_str()) as &str);
 
-    match Operation::from_clap(&app.get_matches()) {
-        Operation::Version { backend, extended, } => {
+    let matches = match app.clone().try_get_matches() {
+        Ok(v) => v,
+        Err(e) => {
+            use clap::error::ErrorKind;
+            return match e.kind() {
+                ErrorKind::DisplayHelp => {
+                    e.exit()
+                },
+                ErrorKind::UnknownArgument =>
+                    Err(anyhow::Error::from(Error::UnsupportedOption)
+                        .context(format!("{}", e))),
+                _ => Err(e.into()),
+            };
+        },
+    };
+
+    match Operation::from_arg_matches(&matches)? {
+        Operation::Version { backend, extended, sop_spec } => {
             let version = sop.version()?;
-            match (backend, extended) {
-                (false, false) => {
+            match (backend, extended, sop_spec) {
+                (false, false, false) => {
                     let v = version.frontend()?;
                     println!("{} {}", v.name, v.version);
                 },
-                (true, false) => {
+                (true, false, false) => {
                     let v = version.backend()?;
                     println!("{} {}", v.name, v.version);
                 },
-                (false, true) => {
-                    let mut versions = version.extended()?;
-                    versions.insert(1, crate::VersionInfo {
-                        name: "sop-rs".into(),
-                        version: env!("CARGO_PKG_VERSION").into(),
-                    });
-                    for v in versions {
-                        println!("{} {}", v.name, v.version);
-                    }
+                (false, true, false) => {
+                    let v = version.frontend()?;
+                    println!("{} {}", v.name, v.version);
+                    println!("sop-rs {}", env!("CARGO_PKG_VERSION"));
+                    println!("{}", version.extended()?);
                 },
-                (true, true) => unreachable!("mutually exclusive"),
+                (false, false, true) => {
+                    println!("{}", sop.spec_version());
+                },
+                _ => unreachable!("flags are mutually exclusive"),
             }
         },
 
-        Operation::GenerateKey { no_armor, with_key_password, userids, } => {
+        Operation::ListProfiles { subcommand, } => {
+            let profiles = match subcommand.as_str() {
+                "generate-key" => sop.generate_key()?.list_profiles(),
+                "encrypt" => sop.encrypt()?.list_profiles(),
+                _ => return Err(Error::UnsupportedProfile.into()),
+            };
+            for (p, d) in profiles {
+                println!("{}: {}", p, d);
+            }
+        },
+
+        Operation::GenerateKey {
+            no_armor,
+            profile,
+            signing_only,
+            with_key_password,
+            userids,
+        } => {
             let mut op = sop.generate_key()?;
-            if no_armor {
-                op = op.no_armor();
+            if let Some(p) = profile {
+                op = op.profile(&p)?;
+            }
+            if signing_only {
+                op = op.signing_only();
             }
             if let Some(p) = with_key_password {
                 op = op.with_key_password(get_password(p)?)?;
@@ -351,15 +468,39 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
             for u in userids {
                 op = op.userid(&u);
             }
-            op.generate()?.write_to(&mut io::stdout())?;
+            op.generate()?
+                .to_writer(! no_armor, &mut io::stdout())?;
         },
 
-        Operation::ExtractCert { no_armor, } => {
-            let mut op = sop.extract_cert()?;
-            if no_armor {
-                op = op.no_armor();
+        Operation::ChangeKeyPassword {
+            no_armor,
+            new_key_password,
+            old_key_password,
+        } => {
+            let mut op = sop.change_key_password()?;
+            if let Some(p) = new_key_password {
+                op = op.new_key_password(get_password(p)?)?;
             }
-            op.keys(&mut io::stdin())?.write_to(&mut io::stdout())?;
+            for p in old_key_password {
+                op = op.old_key_password(get_password(p)?)?;
+            }
+            op.keys(&K::from_reader(sop, &mut io::stdin())?)?
+                .to_writer(! no_armor, &mut io::stdout())?;
+        }
+
+        Operation::RevokeKey { no_armor, with_key_password, } => {
+            let mut op = sop.revoke_key()?;
+            for p in with_key_password {
+                op = op.with_key_password(get_password(p)?)?;
+            }
+            op.keys(&K::from_reader(sop, &mut io::stdin())?)?
+                .to_writer(! no_armor, &mut io::stdout())?;
+        }
+
+        Operation::ExtractCert { no_armor, } => {
+            sop.extract_cert()?
+                .keys(&K::from_reader(sop, &mut io::stdin())?)?
+                .to_writer(! no_armor, &mut io::stdout())?;
         },
 
         Operation::Sign {
@@ -370,21 +511,18 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
             keys,
         } => {
             let mut op = sop.sign()?.mode(as_);
-            if no_armor {
-                op = op.no_armor();
-            }
             for p in with_key_password {
                 op = op.with_key_password(get_password_unchecked(p)?)?;
             }
-            for mut key in load_files(keys)? {
-                op = op.key(&mut key)?;
+            for mut stream in load_files(keys)? {
+                op = op.keys(&K::from_reader(sop, &mut stream)?)?;
             }
-            let micalg =
-                op.data(&mut io::stdin())?.write_to(&mut io::stdout())?;
+            let (micalg, sigs) = op.data(&mut io::stdin())?;
             if let Some(path) = micalg_out {
                 let mut sink = create_file(path)?;
                 write!(sink, "{}", micalg)?;
             }
+            sigs.to_writer(! no_armor, &mut io::stdout())?;
         },
 
         Operation::Verify { not_before, not_after, signatures, certs, } => {
@@ -395,10 +533,11 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
             if let Some(t) = not_after {
                 op = op.not_after(t.into());
             }
-            for mut cert in load_files(certs)? {
-                op = op.certs(&mut cert)?;
+            for mut stream in load_files(certs)? {
+                op = op.certs(&C::from_reader(sop, &mut stream)?)?;
             }
-            let verifications = op.signatures(&mut load_file(signatures)?)?
+            let verifications =
+                op.signatures(&Sigs::from_reader(sop, &mut load_file(signatures)?)?)?
                 .data(&mut io::stdin())?;
 
             for v in verifications {
@@ -408,30 +547,51 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
 
         Operation::Encrypt {
             no_armor,
+            profile,
             as_,
             with_password,
             with_key_password,
             sign_with,
+            session_key_out,
             certs,
         } =>
         {
+            let session_key_out: Option<Box<dyn io::Write>> =
+                if let Some(f) = session_key_out {
+                    Some(Box::new(create_file(f)?))
+                } else {
+                    None
+                };
+
             let mut op = sop.encrypt()?.mode(as_);
             if no_armor {
                 op = op.no_armor();
             }
+            if let Some(p) = profile {
+                op = op.profile(&p)?;
+            }
             for p in with_key_password {
                 op = op.with_key_password(get_password_unchecked(p)?)?;
             }
-            for mut key in load_files(sign_with)? {
-                op = op.sign_with_key(&mut key)?;
+            for mut stream in load_files(sign_with)? {
+                op = op.sign_with_keys(&K::from_reader(sop, &mut stream)?)?;
             }
             for pw in with_password.into_iter().map(get_password) {
                 op = op.with_password(pw?)?;
             }
-            for mut cert in load_files(certs)? {
-                op = op.with_cert(&mut cert)?;
+            for mut stream in load_files(certs)? {
+                op = op.with_certs(&C::from_reader(sop, &mut stream)?)?;
+            }
+            let session_key =
+                op.plaintext(&mut io::stdin())?.to_writer(&mut io::stdout())?;
+
+            if let Some(mut sko) = session_key_out {
+                if let Some(sk) = session_key {
+                    writeln!(sko, "{}", sk)?;
+                } else {
+                    return Err(Error::UnsupportedSubcommand.into());
+                }
             }
-            op.plaintext(&mut io::stdin())?.write_to(&mut io::stdout())?;
         },
 
         Operation::Decrypt {
@@ -481,12 +641,12 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
                 op = op.with_key_password(get_password_unchecked(p)?)?;
             }
 
-            for mut key in load_files(keys)? {
-                op = op.with_keys(&mut key)?;
+            for mut stream in load_files(keys)? {
+                op = op.with_keys(&K::from_reader(sop, &mut stream)?)?;
             }
 
-            for mut cert in load_files(verify_with)? {
-                op = op.verify_with_certs(&mut cert)?;
+            for mut stream in load_files(verify_with)? {
+                op = op.verify_with_certs(&C::from_reader(sop, &mut stream)?)?;
             }
 
             if let Some(t) = verify_not_before {
@@ -498,7 +658,7 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
             }
 
             let (session_key, verifications) =
-                op.ciphertext(&mut io::stdin())?.write_to(&mut io::stdout())?;
+                op.ciphertext(&mut io::stdin())?.to_writer(&mut io::stdout())?;
 
             for v in verifications {
                 writeln!(verify_out, "{}", v)?;
@@ -513,26 +673,24 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
         },
 
         Operation::Armor { label, } => {
+            #[allow(deprecated)]
             let op = sop.armor()?.label(label);
             op.data(&mut io::stdin())?
-                .write_to(&mut io::stdout())?;
+                .to_writer(&mut io::stdout())?;
         },
 
         Operation::Dearmor {} => {
             let op = sop.dearmor()?;
             op.data(&mut io::stdin())?
-                .write_to(&mut io::stdout())?;
+                .to_writer(&mut io::stdout())?;
         },
 
         Operation::InlineDetach { no_armor, signatures_out, } => {
-            let mut op = sop.inline_detach()?;
-            if no_armor {
-                op = op.no_armor();
-            }
+            let op = sop.inline_detach()?;
             let mut signatures_out = create_file(signatures_out)?;
             let signatures =
-                op.message(&mut io::stdin())?.write_to(&mut io::stdout())?;
-            signatures_out.write_all(&signatures)?;
+                op.message(&mut io::stdin())?.to_writer(&mut io::stdout())?;
+            signatures.to_writer(! no_armor, &mut signatures_out)?;
         },
 
         Operation::InlineVerify {
@@ -555,11 +713,11 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
                 } else {
                     Box::new(io::sink())
                 };
-            for mut cert in load_files(certs)? {
-                op = op.certs(&mut cert)?;
+            for mut stream in load_files(certs)? {
+                op = op.certs(&C::from_reader(sop, &mut stream)?)?;
             }
             let verifications =
-                op.message(&mut io::stdin())?.write_to(&mut io::stdout())?;
+                op.message(&mut io::stdin())?.to_writer(&mut io::stdout())?;
 
             for v in verifications {
                 writeln!(&mut verifications_out, "{}", v)?;
@@ -574,10 +732,10 @@ fn real_main(sop: &dyn crate::SOP) -> Result<()> {
             for p in with_key_password {
                 op = op.with_key_password(get_password_unchecked(p)?)?;
             }
-            for mut key in load_files(keys)? {
-                op = op.key(&mut key)?;
+            for mut stream in load_files(keys)? {
+                op = op.keys(&K::from_reader(sop, &mut stream)?)?;
             }
-            op.data(&mut io::stdin())?.write_to(&mut io::stdout())?;
+            op.data(&mut io::stdin())?.to_writer(&mut io::stdout())?;
         },
 
         Operation::Unsupported(args) => {
@@ -722,22 +880,48 @@ fn load_files(files: Vec<String>)
 }
 
 /// Parses the given string depicting a ISO 8601 timestamp, rounding down.
-fn parse_bound_round_down(s: &str) -> Result<DateTime<Utc>> {
-    match s {
-        // XXX: parse "-" to None once we figure out how to do that
-        // with structopt.
-        "now" => Ok(Utc::now()),
-        _ => parse_iso8601(s, NaiveTime::from_hms_opt(0, 0, 0).unwrap()),
+#[derive(Debug, Clone, PartialEq, Eq, Hash)]
+struct NotBefore(DateTime<Utc>);
+
+impl std::str::FromStr for NotBefore {
+    type Err = anyhow::Error;
+
+    fn from_str(s: &str) -> Result<NotBefore> {
+        match s {
+            // XXX: parse "-" to None once we figure out how to do that
+            // with clap.
+            "now" => Ok(Utc::now()),
+            _ => parse_iso8601(s, NaiveTime::from_hms_opt(0, 0, 0).unwrap()),
+        }.map(|t| NotBefore(t))
+    }
+}
+
+impl From<NotBefore> for std::time::SystemTime {
+    fn from(t: NotBefore) -> std::time::SystemTime {
+        t.0.into()
     }
 }
 
 /// Parses the given string depicting a ISO 8601 timestamp, rounding up.
-fn parse_bound_round_up(s: &str) -> Result<DateTime<Utc>> {
-    match s {
-        // XXX: parse "-" to None once we figure out how to do that
-        // with structopt.
-        "now" => Ok(Utc::now()),
-        _ => parse_iso8601(s, NaiveTime::from_hms_opt(23, 59, 59).unwrap()),
+#[derive(Debug, Clone, PartialEq, Eq, Hash)]
+struct NotAfter(DateTime<Utc>);
+
+impl std::str::FromStr for NotAfter {
+    type Err = anyhow::Error;
+
+    fn from_str(s: &str) -> Result<NotAfter> {
+        match s {
+            // XXX: parse "-" to None once we figure out how to do that
+            // with clap.
+            "now" => Ok(Utc::now()),
+            _ => parse_iso8601(s, NaiveTime::from_hms_opt(23, 59, 59).unwrap()),
+        }.map(|t| NotAfter(t))
+    }
+}
+
+impl From<NotAfter> for std::time::SystemTime {
+    fn from(t: NotAfter) -> std::time::SystemTime {
+        t.0.into()
     }
 }
 
@@ -789,7 +973,7 @@ fn parse_iso8601(s: &str, pad_date_with: chrono::NaiveTime)
 
 #[test]
 fn test_parse_iso8601() {
-    let z = chrono::NaiveTime::from_hms(0, 0, 0);
+    let z = chrono::NaiveTime::from_hms_opt(0, 0, 0).unwrap();
     parse_iso8601("2017-03-04T13:25:35Z", z).unwrap();
     parse_iso8601("2017-03-04T13:25:35+08:30", z).unwrap();
     parse_iso8601("2017-03-04T13:25:35", z).unwrap();
@@ -894,6 +1078,12 @@ enum Error {
     #[error("Options were supplied that are incompatible with each other")]
     IncompatibleOptions,
 
+    /// The requested profile is unsupported (sop generate-key, sop
+    /// encrypt), or the indicated subcommand does not accept profiles
+    /// (sop list-profiles).
+    #[error("Profile not supported")]
+    UnsupportedProfile,
+
     /// An I/O operation failed.
     #[error("I/O operation failed")]
     IoError(#[source] io::Error),
@@ -921,6 +1111,7 @@ impl From<crate::Error> for Error {
             E::AmbiguousInput => CE::AmbiguousInput,
             E::IncompatibleOptions => CE::IncompatibleOptions,
             E::NotImplemented => CE::UnsupportedSubcommand,
+            E::UnsupportedProfile => CE::UnsupportedProfile,
             E::IoError(e) => CE::IoError(e),
         }
     }
@@ -948,6 +1139,7 @@ impl From<Error> for i32 {
             AmbiguousInput => 73,
             KeyCannotSign => 79,
             IncompatibleOptions => 83,
+            UnsupportedProfile => 89,
             IoError(_) => 1,
         }
     }
